/**
 * @file main.c
 * @brief Este projeto eh um template para o projeto do roteiro 9 da disciplina EA871 do
 * segundo semestre de 2020
 * @author Wu Shin-Ting
 * @date 06/11/2020
 */

#include "derivative.h" /* include peripheral declarations */

/********************************************************
 * Utilities
 */
#define GPIO_PIN(x) (1<<x)

/**
 * @brief atraso em numero de interacoes <i>
 * @param[in] i numero de iteracoes
 */
void delay (uint32_t i)
{
	__asm("outerLoop:;"
			"sub %0, %0, #1;"
			"mov r2, #11;"
			"innerLoop:;"
			"sub r2, r2, #1;"
			"cmp r2, #0;"
			"bne innerLoop;"
			"cmp %0, #0;"
			"bne outerLoop;"
			: 
			: "r" (i)
	);
}

char *ttoa (unsigned int seconds, char *string)
{
	unsigned int hh, mm, ss;
	
	ss = seconds%60;
	
	mm = (seconds/60)%60;
	
	hh = seconds/3600;
	
	string[2] = string[5] = ':';    // inserir os espacadores
	string[8] = '\0';				// terminador

	if (hh > 23) {
		// horario invalido: FF:FF:FF
		string[0] = string[1] = string[3] = string[4] = string[6] = string[7] = 'F';
	}
	
	string[0] = (hh < 10)? '0': (hh/10)+'0';
	string[1] = hh%10+'0';
	
	string[3] = (mm < 10)? '0': (mm/10)+'0';
	string[4] = mm%10+'0';
	
	string[6] = (ss < 10)? '0': (ss/10)+'0';
	string[7] = ss%10+'0';	
	
	return string;
}

/********************************************************
 * Switches
 */
/**
 * @brief inicializa o componente Botoeiras 
 */
void GPIO_initSwitches () {
	// Configure pinos das botoeiras
	SIM_SCGC5 |= SIM_SCGC5_PORTA_MASK;				// para botoeiras   					// para leds

	//Funcao GPIO
	PORTA_PCR4 &= ~PORT_PCR_MUX(0x7);			// limpar os bits
	PORTA_PCR4 |= PORT_PCR_MUX(0x1);			// antes de setar 0b001

	PORTA_PCR5 |= PORT_PCR_MUX(0x1);

	PORTA_PCR12 |= PORT_PCR_MUX(0x1);

	//Sentido do sinal: entrada
	GPIOA_PDDR &= ~(GPIO_PDDR_PDD(GPIO_PIN(4)) |
			GPIO_PDDR_PDD(GPIO_PIN(5)) |
			GPIO_PDDR_PDD(GPIO_PIN(12)));    
}

/**
 * @brief Habilita interrupcoes da PORTA no NVIC.
 * @param[in] priority level
 */
void GPIO_enableSwInterrupt(uint8_t priority) {
	/* 
	 * "Non-core interrupt source count" (IRQ = Vector Number (46) - 16, 
	 * Table 3-7 de KL-25 Sub-Family Reference Manual) eh habilitado
	 * ao escrever 1 no bit correpondente do registrador NVIC_ISER 
	 * (B3.4.3 de ARMv6-M Architecture Reference Manual) e a solicitacao de 
	 * interrupcao eh limpa ao escrever 1 no bit correspondente do 
	 * registrador NVIC_ICPR (B3.4.6 de ARMv6-M Architecture Reference Manual) 
	 */
	/* PORTA: Vetor 46, IRQ30, prioridade em NVICIPR7[23:16]
	 * Secao 3.3.2.3 em 
	 * <a href="ftp://ftp.dca.fee.unicamp.br/pub/docs/ea871/ARM/KL25P80M48SF0RM.pdf" target=_manual>ftp://ftp.dca.fee.unicamp.br/pub/docs/ea871/ARM/KL25P80M48SF0RM.pdf</a>
	 */
	NVIC_ISER |= 1 << (46-16);            // Habilita interrupcao PORTA
	NVIC_ICPR |= 1 << (46-16);
	NVIC_IPR7 |= NVIC_IP_PRI_30(priority << 6);    // Nivel de prioridade = 3

	// Comfigurar o trigger
	PORTA_PCR4 |= PORT_PCR_ISF_MASK | 
			PORT_PCR_IRQC(0xA);           // "Interrupt on falling edge"
	PORTA_PCR5 |= PORT_PCR_ISF_MASK | 
			PORT_PCR_IRQC(0xA);           // "Interrupt on falling edge"
	PORTA_PCR12 |= PORT_PCR_ISF_MASK | 
			PORT_PCR_IRQC(0xA);            // "Interrupt on falling edge"
}

/********************************************************
 * SysTick
 */
/**
 * @brief Inicializar o temporizador SysTick com um valor de recarga. A base de tempo
 * eh o relogio do nucleo (processador), configurado em 20971529Hz
 * @param[in] periodo valor de recarga
 */
void SysTick_init (unsigned int periodo) {
	SYST_RVR = SysTick_RVR_RELOAD(periodo);   
	SYST_CVR = SysTick_CVR_CURRENT(0);          // reseta flag e recarga     
	SYST_CSR |= (SysTick_CSR_CLKSOURCE_MASK |	// fonte de CLK: nucleo
			SysTick_CSR_TICKINT_MASK);			// habilitar interrupcao
	SYST_CSR &= ~SysTick_CSR_ENABLE_MASK;		// inicializar SysTick desabilitado
}

/**
 * @brief Habilitar SysTick
 */
void SysTick_enable (void) {
	SYST_CSR |= SysTick_CSR_ENABLE_MASK;       // habilitar SysTick 
}

/**
 * @brief Desabilitar SysTick
 */
void SysTick_disable (void) {
	SYST_CSR &= ~SysTick_CSR_ENABLE_MASK;     // desabilitar Systick 
}

/********************************************************
 * LCD
 */
/**
 * @brief Funcao do byte enviado para LCD
 */
typedef enum lcd_RS_tag {
	COMANDO,  /**< Comando */
	DADO	  /**< Dado */
} lcd_RS_type;

/**
 * @brief Inicializa os pinos conectados ao latch 74573 onde estao conectados os 8 leds vermelhos
 */
void GPIO_initLCD(){
	SIM_SCGC5 |= SIM_SCGC5_PORTC_MASK ;     ///< Habilita o clock da PORTA C

	// Configura os pinos PORTC[10:0] como GPIO de saida 
	PORTC_PCR0 |= PORT_PCR_MUX(0x1);       ///< D0-D7 dos dados (GPIO)
	PORTC_PCR1 |= PORT_PCR_MUX(0x1);
	PORTC_PCR2 |= PORT_PCR_MUX(0x1);
	PORTC_PCR3 |= PORT_PCR_MUX(0x1);
	PORTC_PCR4 |= PORT_PCR_MUX(0x1);
	PORTC_PCR5 |= PORT_PCR_MUX(0x1);
	PORTC_PCR6 |= PORT_PCR_MUX(0x1);
	PORTC_PCR7 |= PORT_PCR_MUX(0x1);

	PORTC_PCR8 |= PORT_PCR_MUX(0x1);       ///< RS do LCD
	PORTC_PCR9 |= PORT_PCR_MUX(0x1);       ///< E do LCD

	GPIOC_PDDR |= GPIO_PDDR_PDD(GPIO_PIN(9)|GPIO_PIN(8) |GPIO_PIN(7) | 
			GPIO_PIN(6) | GPIO_PIN(5) | GPIO_PIN(4) | GPIO_PIN(3) | 
			GPIO_PIN(2) | GPIO_PIN(1) | GPIO_PIN(0));       ///< Configura como saidas
	GPIOC_PDOR &= ~GPIO_PDOR_PDO(GPIO_PIN(9) |GPIO_PIN(8)|GPIO_PIN(7) | 
			GPIO_PIN(6) | GPIO_PIN(5) | GPIO_PIN(4) | GPIO_PIN(3) | 
			GPIO_PIN(2) | GPIO_PIN(1) | GPIO_PIN(0)); ///< Inicializa os pinos com 0
}

/**
 * @brief Escrever nos pinos de dados um caractere de 8 bits
 * @param[in] c caractere
 */
void GPIO_setByte(char c) {
	// colocar os dados nos pinos de dados DO-D7
	GPIOC_PDOR &= 0xFFFFFF00; 		// zerar o byte menos significativo
	GPIOC_PDOR |= c; 				// setas os bits em 1 no byte menos significativo	
}

/**
 * @brief Gerar o pulso de Enable de ~450ns
 */
void GPIO_enableE()
{
	// gerar um pulso  (~ 450 ns)
	GPIOC_PSOR = GPIO_PIN (9);		// setar em 1
	delay(1);
	GPIOC_PCOR = GPIO_PIN (9);		// resetar em 0
}

/**
 * @brief Setar o tipo de transferencia
 * @param[in] i tipo
 */
void GPIO_setRS (lcd_RS_type i) {
	if(i == COMANDO) 
		GPIOC_PCOR = GPIO_PIN(8);     // setar o LCD no modo de comando		
	else if (i == DADO) 
		GPIOC_PSOR = GPIO_PIN(8);    // setar o LCD no modo de dados
}

/**
 * @brief Habilitar cursor piscante
 */
void GPIO_enableBlinkedCursor (void) {
	GPIOC_PCOR = GPIO_PIN(8);     	// modo comando		
	GPIOC_PDOR &= ~(0b11111111); 	    // mask	
	GPIOC_PDOR |= (0b00001111); 	// byte de comando	
	GPIOC_PSOR = GPIO_PIN (9);		// pulso E
	delay (1);
	GPIOC_PCOR = GPIO_PIN (9);		// resetar em 0
	delay (20);
}

/**
 * @brief Desabilitar o cursor piscante
 */
void GPIO_disableBlinkedCursor (void) {
	GPIOC_PCOR = GPIO_PIN(8);     	// modo comando		
	GPIOC_PDOR &= ~(0b11111111); 	    // mask	
	GPIOC_PDOR |=  (0b00001100); 	// byte de comando	
	GPIOC_PSOR = GPIO_PIN (9);		// pulso E
	delay(1);
	GPIOC_PCOR = GPIO_PIN (9);		// resetar em 0
	delay (20);
}

/**
 * @brief Transferir um byte ao LCD
 * @param[in] c byte
 * @param[in] t tempo de processamento em multiplos de 2 us
 */
void escreveLCD(char c, uint32_t t)
{
	GPIO_setByte(c);
	GPIO_enableE();
	delay(t);
}

/**
 * @brief Setar o endereco do cursor no visor
 */
void GPIO_posicioneCursor (uint8_t end) {
	GPIO_setRS(COMANDO);			// comando
	escreveLCD(0b10000000|end, 20);			// seta o endereco do cursor
}

/**
 * @brief Inicializar o LCD
 */
void initLCD () {
	//espera por mais de 30ms = 30000us/2 = 15000
	delay(15000);

	GPIO_setRS(COMANDO);
	escreveLCD(0x38, 20);			//Function Set: 39us/2 -> 20
	escreveLCD(0x0C, 20);		    //Display ON/OFF Control: 39us/2 -> 20
	escreveLCD(0x01, 765);			//Display Clear: 1530us/2 -> 765
	escreveLCD(0x06, 20);			//Entry mode set: 39us/2 -> 20
}

/**
 * @brief Limpe o display
 */
void limpeDisplayLCD () {
	GPIO_setRS(COMANDO);
	escreveLCD(0x01, 765);			//Display Clear: 1530us/2 -> 765
}

/**
 * @brief Escrever uma mensagem str no LCD a partir do endereco especificado
 * @param[in] end endereco DDRAM
 * @param[in] str mensagem
 */
void escreveMensagem (uint8_t end, char* str)
{
	//Posicione o cursor no end
	uint8_t tmp = 0b10000000 | end; // seta o bit 7 em 1
	GPIO_setRS(COMANDO);			// comando
	escreveLCD(tmp, 20);			// seta o endereco do cursor

	//Chaveia para modo DADO
	GPIO_setRS (DADO);
	while (*str) {
		escreveLCD(*str, 20);
		str++;
	}
}

/**
 * @brief Escrever um bitmap na memoria CGRAM a partir do endereco especificado
 * @param[in] end endereco CGRAM
 * @param[in] str bitmap
 */
void escreveBitmap (uint8_t end, char* bitmap)
{
	//Posicione o cursor no end
	uint8_t tmp = 0b01000000 | (end*8); // seta o bit 7 em 0 e bit 6 em 1
	GPIO_setRS(COMANDO);				// comando
	escreveLCD(tmp, 20);				// seta o endereco do cursor

	//Chaveia para modo DADO
	GPIO_setRS (DADO);
	uint8_t i;
	for (i=0; i< 8; i++) {
		escreveLCD(bitmap[i], 20);
	}
}

/********************************************************
 * RTC - 1kHz
 */
/**
 * @brief Inicializar o modulo RTC
 */
void RTC_init()
{
	SIM_SCGC6 |= SIM_SCGC6_RTC_MASK;	// habilitar sinal CLK para RTC
	
	SIM_SOPT1 |= SIM_SOPT1_OSC32KSEL(0b11);  // setar LPO 1kHz como fonte de ERCLK32K
	
	//Ao inves do prescaler (RTC_TPR) incrementar a 32.768kHz, incrementara a 1kHz 

	RTC_CR |= RTC_CR_SWR_MASK;			// resetar os registradores do RTC por software
										// limpar o contador RTC_TSR e flags RTC_SR_TOF e RTC_SR_TIF
	RTC_CR &= ~RTC_CR_SWR_MASK;			// necessario resetar explicitamente SWR 
	
    RTC_TCR = RTC_TCR_CIR(0x00) |		// ajustar o erro de frequencia (ppm) em cada segundo
              RTC_TCR_TCR(0x00);        // 1s = 32.768 ticks (sem ajuste)
    
    RTC_TPR = 0; 			// resetar prescaler (2^15 - 1)
    RTC_TSR = 0;			// resetar o contador para que a bandeira 
    						// TIF seja baixada.
    
	RTC_IER &= ~(RTC_IER_TAIE_MASK |	// desabilitar a interrupcao por alarme
				RTC_IER_TOIE_MASK  |	// desabilitar a interrupcao por overflow
				RTC_IER_TSIE_MASK);    	// desabilitar a interrupcao por segundo

	RTC_SR |= RTC_SR_TCE_MASK;			// habilita incremento de Prescaler
}

/**
 * @brief Habilitar o tratamento de interrupcao por alarme no NVIC
 */
void RTC_enableNVICAlarm (char priority) {

    NVIC_ISER |= 1 << (36-16);            ///< habilitar IRQ 20
    NVIC_ICPR |= 1 << (36-16);
    NVIC_IPR5 |= NVIC_IP_PRI_20(priority << 6);   // 20/4
}

/**
 * @brief Habilitar o tratamento de interrupcao por segundo no NVIC
 */
void RTC_enableNVICSecond (char priority) {

    NVIC_ISER |= 1 << (37-16);            ///< habilitar IRQ 21
    NVIC_ICPR |= 1 << (37-16);
    NVIC_IPR5 |= NVIC_IP_PRI_21(priority << 6);   // 21/4
}

/**
 * @brief Habilitar interrupcao por alarme
 */
void RTC_enableInterruptAlarm (void) {
	RTC_IER |= RTC_IER_TAIE_MASK;    	// habilitar a interrupcao por segundo
}

/**
 * @brief Desabilitar a interrupcao por alarme
 */
void RTC_disableInterruptAlarm(void) {
	RTC_IER &= ~RTC_IER_TAIE_MASK;    // desabilitar a interrupcao por segundo
}

/**
 * @brief Habilitar interrupcao por segundo
 */
void RTC_enableInterruptSecond (void) {
	RTC_IER |= RTC_IER_TSIE_MASK;    	// habilitar a interrupcao por segundo
}

/**
 * @brief Desabilitar a interrupcao por segundo
 */
void RTC_disableInterruptSecond(void) {
	RTC_IER &= ~RTC_IER_TSIE_MASK;    // desabilitar a interrupcao por segundo
}

/**
 * @brief Ler o valor atual do contador de segundos do RTC
 * @param[out] seconds medida de tempo em segundos
 */
void RTC_getTime (unsigned int *seconds) {
	*seconds = (RTC_TSR*32768)/1000+(RTC_TPR/1000);    // reajustar valor para segundos
}

/**
 * @brief Setar o valor do contador de segundos do RTC
 * @param[in] seconds valor de tempo em segundos
 */
void RTC_setTime (unsigned int seconds) {
	RTC_SR &= ~RTC_SR_TCE_MASK;			// desabilita o modulo para atualizar prescaler
										// e contador
	RTC_TPR = (seconds*1000)%32768;     // reajustar segundos para a frequencia 1kHz
	RTC_TSR = (seconds*1000)/32768;		// reajustar segundos para a frequencia 1kHz
	RTC_SR |= RTC_SR_TCE_MASK;			// habilitar o modulo
}

/**
 * @brief Setar o valor do alarme do RTC 
 * @param[in] seconds valor de alarme em segundos
 */
void RTC_setAlarm (unsigned int seconds) {
	RTC_TAR = (seconds*1000)/32768;		// reajustar segundos para a frequencia 1kHz
}

/********************************************************
 * Application
 */
/**
 * @brief Modo do relogio
 */
typedef enum rtc_mode_tag {
	NORMAL,   /**< Operacao normal (0) */
	AJUSTE	  /**< Ajuste (1) */
} rtc_mode_type;

/**
 * @brief Unidades do relogio
 */
typedef enum rtc_unit_tag {
	HH,   /**< Horas (0)*/
	MM,	  /**< Minutos (1) */
	SS	  /**< Segundos (2) */
} rtc_unit_type;

rtc_mode_type mode;
uint8_t units;
uint8_t flagP, flagH;      // resetar flag de controle da atualizacao do visor do LCD
unsigned int s_cur = 0;
unsigned short horario[3]; // 0 (HH): horas; 1(MM): minutos; 2 (SS): segundos
unsigned short ciclos_completos;

/**
 * Rotina de servico para SysTick: controlar Timeout
 */
void SysTick_Handler () {
	ciclos_completos++;

	if (ciclos_completos > 4) {
		//Timeout
		mode = NORMAL;				// retornar para modo NORMAL
		GPIO_disableBlinkedCursor(); // desabilitat o cursor piscante
		//setar novo horario no RTC com o conteudo do vetor horario
		s_cur++;
		SysTick_disable();
	}	
}

/**
 * @brief Rotina de servico para PORTA (Vetor 46, IRQ30, prioridade em NVICIPR7[31:30])
 */
void PORTA_IRQHandler(void) {

	// iniciar contagem para Timeout
	ciclos_completos = 0;
	SysTick_enable();

	if (PORTA_PCR4 & PORT_PCR_ISF_MASK) {
		// decrementar o conteudo de horario[units]
		flagH = 1;
		
		// limpar flag
		PORTA_PCR4 |= PORT_PCR_ISF_MASK; 
		return;
	}

	if (PORTA_PCR5 & PORT_PCR_ISF_MASK) {
		// incrementar o conteudo de horario[units]
		flagH = 1;
		
		// limpar flag
		PORTA_PCR5 |= PORT_PCR_ISF_MASK; 
		return;
	} 

	if (PORTA_PCR12 & PORT_PCR_ISF_MASK) {
		flagP = 1;
		if (mode == NORMAL) {
			mode = AJUSTE;
			units = HH;
			GPIO_enableBlinkedCursor();
		} else {
			units++;
			units = units%3;   
		}
		// limpa flag
		PORTA_PCR12 |= PORT_PCR_ISF_MASK; 
		return;
	}
}

/**
 * @brief Fluxo de controle principal do relogio digital 
 */
int main(void)
{
	GPIO_initLCD();				// Inicializar pinos de controle LCD
	initLCD();					// Inicializar o LCD

	SysTick_init(10485760);  	// Intervalo de contagem: 10485760/20971520 = 0.5s 	

	GPIO_initSwitches();		// Inicializar pinos de controle de chaves
	GPIO_enableSwInterrupt(1);	// Habilitar interrupcao das chaves

	mode = NORMAL;
	unsigned int s_pre;

	s_pre = s_cur;
	
	s_cur++;
	
	for(;;) {	   
		switch (mode) {
		case NORMAL:
			///mostrar horario no formato HH:MM:SS
			if (s_pre != s_cur)
				escreveMensagem (0x40, "HH:MM:SS");
			///atualizar o conteudo do vetor horario com valor do RTC
			break;
		case AJUSTE:
			if (flagP) {
				if (units == HH){
					// posicione o curso no campo HH
					GPIO_posicioneCursor (0x41);
				} else if (units == MM) {
					// posicione o cursor no campo MM
					GPIO_posicioneCursor (0x44);
				} else if (units == SS) {
					// posicione o cursor no campo SS
					GPIO_posicioneCursor (0x47);
				}
				flagP = 0;
			} else if (flagH) {
				if (units == HH){
				// atualizar o conteudo horario[HH]
					escreveMensagem (0x40, "AA");
				} else if (units == MM){
				// atualizar o conteudo horario[MM]
					escreveMensagem (0x43, "AA");
				} else if (units == SS){
				// atualizar o conteudo horario[SS]
					escreveMensagem (0x46, "AA");
				}
				flagH = 0;
				
			}
			break;
		}
	}

	return 0;
}
